# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2016 Continuum Analytics, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
from click.testing import CliRunner
import pytest

from .base import VCR
from anaconda_enterprise.cli.main import group


@pytest.mark.integration
def test_create():
    runner = CliRunner()
    with VCR.use_cassette('login.yml'):
        result = runner.invoke(group, ['login'], 'derek\nderek')
        assert result.exit_code == 0


#     with VCR.use_cassette('channels/create.yml'):
#         result = runner.invoke(
#             group,
#             ['channels', 'create', 'test1'], )
#         assert result.output == '''\
# test1
#     '''
#         assert result.exit_code == 0



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2016 Continuum Analytics, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
import click

from anaconda_enterprise.cli.api import _list, create_client, pass_client


@click.group(name='channels')
@click.pass_context
def main(ctx):
    """Manage conda channels"""
    ctx.obj = create_client(ctx.obj)


def _list_channels(channels):
    _list(('Name', 'Updated (UTC)'), ('name', 'updated'),
          [c['attributes'] for c in channels])


@main.command()
@pass_client
def list(client):
    _list_channels(client.channels())


@main.command()
@click.argument('name')
@pass_client
def create(client, name):
    _list_channels([client.create_channel(name)])


def _get_namespace(group, user, role):
    if group:
        return group, 'groups'
    elif user:
        return user, 'users'
    elif role:
        return role, 'roles'
    else:
        return None, None


@main.command()
@click.argument('channel_name', metavar='channel')
@click.option('--group', help='Group to share with')
@click.option('--user', help='User to share with')
@click.option('--role', help='Role to share with')
@click.option('--remove',
              is_flag=True,
              default=False,
              help='Remove share for a given user, group, or role.')
@click.option('--level',
              help='The level of permission to grant to the shared channel',
              type=click.Choice([
                  'r',
                  'rw',
              ]),
              default='r')
@pass_client
def share(client, channel_name, group, user, role, remove, level):
    channel = client.get_channel(channel_name)
    if channel is None:
        click.echo('Error: invalid channel {0}.'.format(channel_name))
        return

    name, namespace = _get_namespace(group, user, role)
    if namespace is not None:
        if remove:
            client.remove_shared_channel(channel_name, name, namespace)
            click.echo('Channel sharing removed.')
        else:
            client.share_channel(channel, name, namespace, level)
            click.echo('Channel shared.')


@main.command()
@click.argument('channel_name', metavar='channel')
@pass_client
def delete(client, channel_name):
    channel = client.get_channel(channel_name)
    if channel is None:
        click.echo('Error: invalid channel {0}.'.format(channel_name))
        return

    client.delete_channel(channel)
    click.echo('Channel "{0}" deleted.'.format(channel_name))



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2016 Continuum Analytics, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
import os.path
import glob

from click.testing import CliRunner

from .base import FIXTURES

from anaconda_enterprise.cli.internal import yaml
from anaconda_enterprise.cli.main import group


def test_import_with_dry_run():
    """This test will not upload anything as there are files in the fixture,
    but it will generate a yaml which can then be checked.
    """
    runner = CliRunner()

    dumppath = os.path.join(FIXTURES, 'site-dump')

    users = glob.glob(os.path.join(dumppath, "*"))

    result = runner.invoke(group, ['admin', 'import'] + users + ['--dry-run', 'True'])
    assert result.exit_code == 0

    data = yaml.load(result.output)
    assert data
    assert len(data["accounts"]) == 3



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2016 Continuum Analytics, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
import datetime
import json
import os
import stat
import sys
import tarfile
import time
from urllib.parse import quote_plus, urljoin, urlparse
from uuid import uuid4

import click
import requests
import requests.packages.urllib3
from requests import Session
from requests.auth import AuthBase
from requests.exceptions import ConnectionError, SSLError
from requests.utils import guess_filename

from anaconda_enterprise.cli.appdirs import AppDirs, EnvAppDirs
from anaconda_enterprise.cli.internal import (ClientError, JSONAPIError,
                                              ServerError)
from anaconda_enterprise.cli.multipart import MultipartEncoder

# Hide warnings when ssl_verify is set to false
requests.packages.urllib3.disable_warnings(
    requests.packages.urllib3.exceptions.InsecureRequestWarning)

JSON_API_MEDIA_TYPE = 'application/vnd.api+json'
EPOCH = datetime.datetime(1970, 1, 1)
HAS_FCHMOD = hasattr(os, 'fchmod')
RETRIES = 5


def _list(headers, keys, content, action=None, what=None, plural=None):
    lens = {}
    _format = []
    if content and not isinstance(content[0], dict):
        content = [dict(zip(keys, c)) for c in content]
    for header, key in zip(headers, keys):
        lens[key] = max(0, len(header), *(len(c[key]) for c in content))
        _format.append('{{{}:{}}}'.format(key, lens[key]))
    _format = '  '.join(_format)
    click.echo(_format.format(**{k: v for k, v in zip(keys, headers)}))
    click.echo(_format.format(**{k: '=' * v for k, v in lens.items()}))
    for line in content:
        click.echo(_format.format(**line))
    if what is not None:
        _file_action(len(content), action=action, what=what, plural=plural)


def _file_action(count, action=None, what=None, plural=None):
    click.echo('{} {}{} {}.'.format(count, what or 'file',
                                    '' if count == 1 else (plural or 's'),
                                    action or 'found'))


def _get_binstar_token_directory():
    if 'BINSTAR_CONFIG_DIR' in os.environ:
        return EnvAppDirs('binstar', 'ContinuumIO',
                          os.environ[str('BINSTAR_CONFIG_DIR')]).user_data_dir
    else:
        return AppDirs('binstar', 'ContinuumIO').user_data_dir


class TokenStore(object):
    def __init__(self, path):
        self.path = path
        try:
            with open(path, 'rb') as fd:
                self.data = json.loads(fd.read().decode('utf-8'))
        except (EnvironmentError, ValueError):
            self.data = {}

    def save(self):
        try:
            # exist_ok=True only exists on python3
            os.makedirs(os.path.dirname(self.path))
        except EnvironmentError:
            pass
        with open(self.path, 'wb') as fd:
            # Make sure that the file has the correct permissions
            if HAS_FCHMOD:
                os.fchmod(fd.fileno(), stat.S_IRUSR | stat.S_IWUSR)
            fd.write(
                json.dumps(self.data, indent=2,
                           sort_keys=True).encode('utf-8'))

        self._save_binstar_tokens()

    def _save_binstar_tokens(self):
        """Save tokens to conda 4.3-compatible location
        """
        binstar_dir = _get_binstar_token_directory()
        try:
            # exist_ok=True only exists on python3
            os.makedirs(binstar_dir)
        except EnvironmentError:
            pass
        for url, session in self.data.items():
            url_safe = quote_plus(url)
            try:
                token = session['bearer_token']
                path = os.path.join(binstar_dir, url_safe + '.token')
                with open(path, 'wb') as fd:
                    # Make sure that the file has the correct permissions
                    if HAS_FCHMOD:
                        os.fchmod(fd.fileno(), stat.S_IRUSR | stat.S_IWUSR)
                    fd.write(token.encode('utf-8'))
            except KeyError:
                continue

    def delete_binstar_tokens(self):
        """Remove tokens from conda 4.3-compatible location."""
        binstar_dir = _get_binstar_token_directory()
        for url, session in self.data.items():
            path = os.path.join(binstar_dir, quote_plus(url) + '.token')
            if os.path.exists(path):
                os.remove(path)

    def set_session(self, url, session):
        self.data[url] = session

    def get_session(self, url):
        session = None
        for domain in sorted(self.data, key=len, reverse=True):
            if url.startswith(domain):
                session = self.data[domain]
                if 'bearer_token' not in session:
                    del self.data[domain]
                    self.save()
                    session = None
                break

        return session


class BearerAuthentication(AuthBase):
    def __init__(self, token_store, url):
        self.token_store = token_store
        self.url = url

    def __call__(self, r):
        session = self.token_store.get_session(self.url)
        if session:
            r.headers['Authorization'] = 'Bearer {0}'.format(
                session['bearer_token'])
        return r


class JSONAPILink(object):
    def __init__(self, url, meta):
        self.url = url
        self.meta = meta

    @staticmethod
    def parse(link):
        if isinstance(link, dict):
            return JSONAPILink(link['url'], link.get('meta', {}))
        else:
            return JSONAPILink(link, {})


class ApiClient(object):
    def __init__(self, config, token_store):
        self.url = config['url'].rstrip('/')
        self.url_s = self.url + '/'
        self.token_store = token_store
        self.session = Session()
        self.session.verify = config['ssl_verify']
        self.session.auth = BearerAuthentication(token_store, self.url)
        self.session.headers['Connection'] = 'Keep-Alive'
        self.session.headers['Content-Type'] = JSON_API_MEDIA_TYPE
        self.session.headers['Accept'] = JSON_API_MEDIA_TYPE
        self.file_session = Session()
        self.file_session.verify = config['ssl_verify']
        self.file_session.headers['Connection'] = 'Keep-Alive'
        self.links = {}

    def make_absolute(self, url):
        relative = not urlparse(url).netloc
        if relative:
            url = urljoin(self.url, url)
        return url

    def _request(self, url, *args, **kwargs):
        method = kwargs.pop('method', 'GET')
        if not url:
            url = self.url
            session = self.session
        elif not ('.' in urlparse(url).netloc or 'headers' in kwargs):
            url = urljoin(self.url_s, url)
            session = self.session
        else:
            # If an external URL is used or headers are supplied
            # we do not transmit our bearer token
            session = self.file_session
        for attempt in range(RETRIES):
            try:
                response = session.request(method, url, *args, **kwargs)
                response.raise_for_status()
            except (SSLError, ConnectionError) as exc:
                what = 'SSL' if isinstance(exc, SSLError) else 'connection'
                if attempt == RETRIES - 1:
                    msg = 'Persistent {} error:\n- {} {}\n- ssl_verify: {}\n- exception: {}'
                    raise ServerError(
                        msg.format(what, method, url, self.session.verify,
                                   exc))
                if attempt == 0:
                    sys.stderr.write('- {} {}\n'.format(method, url))
                msg = 'Warning: transient {} error encountered; retrying ({}/{})\n'
                sys.stderr.write(msg.format(what, attempt + 1, RETRIES))
                time.sleep(0.5 * (attempt + 1))
                continue
            except requests.HTTPError as exc:
                msg = 'Unexpected server error {} ({})\n- {} {}'
                msg = msg.format(exc.response.status_code, exc.response.reason,
                                 method, url)
                try:
                    detail = exc.response.json()['errors'][0]['detail']
                    if detail.startswith('HTTP {}:'.format(
                            exc.response.status_code)):
                        detail = detail.split(':', 1)[1].lstrip()
                    msg += '\n' + detail.capitalize()
                except Exception:
                    pass
                raise ServerError(msg)
            except Exception as exc:
                msg = 'Unexpected operation error:\n- {} {}\n- exception: {}'
                raise ServerError(msg.format(method, url, exc))
            if attempt:
                sys.stderr.write(
                    'Retry loop succeeded; no further action needed.\n')
            return response

    def _json_api(self, url, *args, **kwargs):
        field = kwargs.pop('field', 'data')
        if 'data' in kwargs:
            kwargs['data'] = json.dumps(kwargs['data']).encode('utf-8')
        response = self._request(url, *args, **kwargs)
        return self._parse_json_api(response, field=field)

    def _parse_json_api(self, response, field='field'):
        if response.status_code in (404, 204):
            return None

        try:
            body = response.json()
        except (json.decoder.JSONDecodeError) as exc:
            msg = 'Error decoding JSON from API server:\n- exception: {}\n- url: {}'
            raise ServerError(msg.format(exc, response.url))

        if not any(x in body for x in ('data', 'errors', 'meta')):
            raise ServerError(
                'Non-conformant JSONAPI response (missing expected data, meta, or links)'
            )

        if 'errors' in body:
            raise JSONAPIError(body['errors'])

        if 'links' in body:
            try:
                body['links'] = {
                    k: JSONAPILink.parse(v)
                    for k, v in body['links'].items()
                }
            except Exception as ex:
                msg = ('Non-conformant JSONAPI response: invalid links:\n'
                       '- exception: {}\n- value: {}')
                raise ServerError(msg.format(ex, body['links']))

        if 'data' in body:
            data = body['data']
            resources = data if isinstance(data, list) else [data]
            if any('id' not in resource or 'type' not in resource
                   for resource in resources):
                msg = 'Non-conformant JSONAPI response: invalid resources:\n- value: {}'
                raise ServerError(msg.format(data))

        if field == 'body':
            return body
        if field not in body:
            raise ServerError(
                'Non-conformant JSONAPI response: missing field "{}"'.format(
                    field))
        return body[field]

    def login(self, username, password):
        url = self.links['session'].url
        meta = self._json_api(url,
                              method='POST',
                              auth=(username, password),
                              field='meta')
        session = {'bearer_token': meta['bearer_token']}
        self.token_store.set_session(self.url, session)
        self.token_store.save()

    def check(self):
        self.links = self._json_api('', field='links')

    def channels(self):
        return self._json_api('channels')

    def create_channel(self, name):
        resource = {
            "data": {
                "type": "channel",
                "attributes": {
                    "name": name,
                }
            }
        }
        return self._json_api('channels', method='POST', data=resource)

    def get_channel(self, name):
        for chan in self.channels():
            if chan['attributes']['name'] == name:
                return chan
        return None

    def get_channel_acls(self, name):
        channel = self.get_channel(name)
        if channel is not None:
            return self._json_api('channels/{id}/acl'.format(id=channel['id']))
        return None

    def share_channel(self, channel, name, namespace, permission):
        self._json_api('acls',
                       method='POST',
                       data={
                           'data': {
                               'type': 'acl',
                               'attributes': {
                                   'name': name,
                                   'namespace': namespace,
                                   'permission': permission,
                               },
                               'relationships': {
                                   'channel': {
                                       'data': {
                                           'type': 'channel',
                                           'id': channel['id'],
                                       }
                                   }
                               }
                           },
                       })

    def find_channel_acl(self, channel_name, name, namespace):
        acls = self.get_channel_acls(channel_name)
        for acl in acls:
            if acl['attributes']['name'] == name and acl['attributes'][
                    'namespace'] == namespace:
                return acl
        return None

    def remove_shared_channel(self, channel_name, name, namespace):
        channel_acl = self.find_channel_acl(channel_name, name, namespace)
        if channel_acl is None:
            raise ClientError(
                "Channel ACL not found for {channel_name}, {name}, {namespace}"
                .format(channel_name=channel_name,
                        name=name,
                        namespace=namespace))
        self._json_api('acls/{id}'.format(id=channel_acl['id']),
                       method='DELETE')

    def delete_channel(self, channel):
        self._json_api('channels/{id}'.format(id=channel['id']),
                       method='DELETE')

    def stage_file(self, channel, file_info):
        channel_id = channel['id']
        resource = {
            "data": {
                "type": "upload",
                "attributes": file_info,
                "relationships": {
                    "channel": {
                        "data": {
                            "type": "channel",
                            "id": channel_id,
                        }
                    }
                }
            }
        }
        return self._json_api('uploads', method='POST', data=resource)

    def upload_file(self, upload_info, fd):
        upload = JSONAPILink.parse(upload_info['links']['upload'])
        commit = JSONAPILink.parse(upload_info['links']['commit'])

        url = upload.url
        fields = upload.meta.get('fields', {})
        headers = upload.meta.get('headers', {})
        authorization = upload.meta.get('authorization', 'none')
        if authorization == 'bearer-token':
            headers['Authorization'] = 'Bearer {0}'.format(
                self.token_store.get_session(self.url)['bearer_token'])

        fields = list(fields.items())
        response = self._upload_file(fd, fields, headers, url)

        if 200 <= response.status_code < 300:
            return self._json_api(commit.url, method='POST')
        else:
            raise ServerError(
                'Failed to upload file (server responded with {0}:{1})'.format(
                    response.status_code,
                    response.content,
                ))

    def upload_icon(self, upload_info, pkg):
        """
        Upload package icon file to the object store if it exists.
        """
        upload = JSONAPILink.parse(upload_info['links']['icon-upload'])
        url = upload.url
        fields = upload.meta.get('fields', {})
        fields = list(fields.items())

        headers = upload.meta.get('headers', {})
        authorization = upload.meta.get('authorization', 'none')
        if authorization == 'bearer-token':
            headers['Authorization'] = 'Bearer {0}'.format(
                self.token_store.get_session(self.url)['bearer_token'])

        response = None
        with tarfile.open(pkg, mode="r|bz2") as tf:
            for info in tf:
                if info.name == 'info/icon.png':
                    icon_fd = tf.extractfile(info)
                    response = self._upload_file(
                        WrapperFile(icon_fd, info.size), fields, headers, url)

        if response and (response.status_code < 200
                         or response.status_code >= 300):
            raise ServerError(
                'Failed to upload icon (server responded with {0}:{1})'.format(
                    response.status_code,
                    response.content,
                ))

    def _upload_file(self, fd, fields, headers, url):
        """Upload file to s3."""
        filename = guess_filename(fd) or 'file'
        file_field = ('file', (filename, fd))
        fields.append(file_field)

        encoder = MultipartEncoder(fields)
        headers['Content-Type'] = encoder.content_type

        return self._request(url, method='POST', data=encoder, headers=headers)

    def list_files(self, package):
        return self._json_api('packages/{id}/files'.format(id=package['id']),
                              method='GET')

    def delete_file(self, file_id):
        return self._json_api('files/{id}'.format(id=file_id), method='DELETE')

    def get_file(self, package, subdir, filename):
        all_files = self._json_api(
            'packages/{id}/files'.format(id=package['id']), method='GET')
        # Iterate through all files inside the package and find the right one.
        matches = [
            file_ for file_ in all_files
            if (file_['attributes']['filename'] == filename
                and file_['attributes']['subdir'] == subdir)
        ]

        if not matches:
            package_name = package['attributes']['name']
            msg = 'File \'{0}\' was not found for package \'{1}\''
            raise ServerError(msg.format(filename, package_name))
        elif len(matches) > 1:
            msg = 'More than one file was found'
            raise ServerError(msg)
        else:
            return matches.pop()

    def delete_package(self, package, subdir=None, filename=None):
        if subdir and subdir not in package['attributes']['subdirs']:
            raise ServerError("Package does not have {}".format(subdir))

        if subdir and filename:
            file_ = self.get_file(package, subdir, filename)
            return self._json_api('files/{id}'.format(id=file_['id']),
                                  method='DELETE')

        return self._json_api('packages/{id}'.format(id=package['id']),
                              method='DELETE')

    def packages(self):
        return self._json_api('packages')

    def get_package(self, channel, package):
        # TODO replace this with a better search
        for pkg in self.packages():
            if pkg['attributes']['name'] == package and \
               pkg['attributes']['channel-name'] == channel:
                return pkg
        raise ServerError('Package "{}/{}" was not found'.format(
            channel, package))

    def copy_pkg_file(self, channel_src, subdir, filename, channel_dst_id):
        url = 'channels/{0}'.format(channel_dst_id)
        resource = {
            "data": {
                "source": {
                    "channel": channel_src,
                    "subdir": subdir,
                    "filename": filename,
                }
            }
        }
        return self._json_api(url, data=resource, method='PUT')

    def copy_pkg(self, channel_src, pkg_src, channel_dst_id):
        url = 'channels/{0}'.format(channel_dst_id)
        resource = {
            "data": {
                "source": {
                    "channel": channel_src,
                    "package": pkg_src,
                }
            }
        }
        return self._json_api(url, data=resource, method='PUT')

    def create_pypi_index(self, name):
        resource = {
            "data": {
                "type": "pypi_index",
                "attributes": {
                    "name": name,
                }
            }
        }
        return self._json_api('pypi_indexes', method='POST', data=resource)

    def update_pypi_index(self, pypi_index, attributes):
        resource = {
            "data": {
                "type": "pypi_index",
                "attributes": attributes,
                "id": pypi_index["id"]
            }
        }
        return self._json_api('pypi_indexes/{id}'.format(id=pypi_index['id']),
                              method='PATCH',
                              data=resource)

    def pypi_indexes(self):
        return self._json_api('pypi_indexes', method='GET')

    def get_pypi_index(self, name):
        for index in self.pypi_indexes():
            if index['attributes']['name'] == name:
                return index

        raise ClientError("PyPI index named '{}' not found".format(name))

    def get_pypi_index_acls(self, name):
        pypi_index = self.get_pypi_index(name)
        return self._json_api(
            'pypi_indexes/{id}/acl'.format(id=pypi_index['id']))

    def create_pypi_index_acl(self, pypi_index, name, namespace, permission):
        return self._json_api('pypi_acls',
                              method='POST',
                              data={
                                  'data': {
                                      'type': 'pypi_index_acl',
                                      'attributes': {
                                          'name': name,
                                          'namespace': namespace,
                                          'permission': permission,
                                      },
                                      'relationships': {
                                          'pypi-index': {
                                              'data': {
                                                  'type': 'pypi_index',
                                                  'id': pypi_index['id'],
                                              }
                                          }
                                      }
                                  },
                              })

    def get_pypi_index_acl(self, pypi_index_name, name, namespace):
        acls = self.get_pypi_index_acls(pypi_index_name)
        for acl in acls:
            if acl['attributes']['name'] == name and acl['attributes'][
                    'namespace'] == namespace:
                return acl

        raise ClientError("PyPI index ACL not found for {}, {}, {}".format(
            pypi_index_name, name, namespace))

    def get_pypi_packages(self, pypi_index=None):
        if pypi_index is None:
            url = 'pypi_packages'
        else:
            url = 'pypi_indexes/{id}/packages'.format(id=pypi_index['id'])
        return self._json_api(url)

    def delete_pypi_package(self, package):
        return self._json_api('pypi_packages/{id}'.format(id=package['id']),
                              method='DELETE')

    def delete_pypi_file(self, package_file):
        return self._json_api('pypi_files/{id}'.format(id=package_file['id']),
                              method='DELETE')

    def delete_pypi_index_acl(self, pypi_index_acl):
        return self._json_api('pypi_acls/{id}'.format(id=pypi_index_acl['id']),
                              method='DELETE')

    def delete_pypi_index(self, pypi_index):
        return self._json_api('pypi_indexes/{id}'.format(id=pypi_index['id']),
                              method='DELETE')

    def get_pypi_files(self, pypi_package=None):
        if pypi_package is None:
            url = 'pypi_files'
        else:
            url = 'pypi_packages/{id}/files'.format(id=pypi_package['id'])
        return self._json_api(url)

    def stage_pypi_file(self, pypi_index, file_info):
        resource = {
            "data": {
                "type": "pypi_package_upload",
                "attributes": file_info,
                "relationships": {
                    "pypi-index": {
                        "data": {
                            "type": "pypi_index",
                            "id": pypi_index['id'],
                        }
                    }
                }
            }
        }
        return self._json_api('pypi_uploads', method='POST', data=resource)

    def put_global_file(self, file_path, fd, metadata=None):
        root = self.links['global-files'].url
        url = root + file_path
        metadata = '{}' if metadata is None else json.dumps(metadata)
        headers = self.session.headers.copy()
        headers.update(Metadata=metadata)

        response = self._request(url,
                                 method='PUT',
                                 headers=headers,
                                 allow_redirects=False)
        if 200 <= response.status_code < 300:
            return None
        elif 300 <= response.status_code < 400:
            if 'Location' not in response.headers:
                raise ServerError(
                    'Server redirected without specifying a Location header')
            self._request(response.headers['Location'], method='PUT', data=fd)
        else:
            return self._parse_json_api(response)

    def put_public_file(self, file_path, fd, metadata=None):
        """
        Upload a file into the world-accessible special area of Storage

        :param file_path: real local file
        :param metadata: dict of data, like absolute URLs of contained packages
        :param fd: open file pointing to the
        :return: dict, like new database entry
        """
        root = self.links['public-files'].url
        url = root + file_path
        metadata = '{}' if metadata is None else json.dumps(metadata)
        headers = self.session.headers.copy()
        headers.update(Metadata=metadata)

        response = self._request(url,
                                 method='PUT',
                                 headers=headers,
                                 allow_redirects=False)
        if 200 <= response.status_code < 300:
            return None
        elif 300 <= response.status_code < 400:
            if 'Location' not in response.headers:
                raise ServerError(
                    'Server redirected without specifying a Location header')
            self._request(response.headers['Location'], method='PUT', data=fd)
        else:
            return self._parse_json_api(response)

    def stage_installer(self, file_info, env_spec_id=None):
        resource = {
            "data": {
                "type": "installer-upload",
                "attributes": file_info,
            }
        }

        if env_spec_id is not None:
            resource['data']['relationships'] = {
                "spec": {
                    "data": {
                        "type": "env_spec",
                        "id": env_spec_id
                    }
                }
            }

        return self._json_api('installer_uploads',
                              method='POST',
                              data=resource)

    def fetch_resource(self, full_url):
        return self._json_api(full_url)

    def store_env_spec_lock(self, env_spec_id, locked_env_specs, meta):
        """
        Creates env spec lock.
        """
        def create_pkg_for_wire(lock_id, pkg, platform):
            """
            Creates the structure required by the handlers/jsonserializer to parse a package
            """
            return {
                'type': 'env_spec_package',
                'id': uuid4().hex,
                'attributes': {
                    'env-spec-lock-id': lock_id,
                    'name': pkg,
                    'platform': platform,
                }
            }

        def create_lock_for_wire(name, lock_id, hash_, packages):
            return {
                'type': 'env_spec_lock',
                'id': lock_id,
                'attributes': {
                    'name': name,
                    'hash': hash_,
                    'spec-id': env_spec_id,  # define as arg of outerfunction
                    'extra-attributes': meta
                },
                'relationships': {
                    'env-spec-packages': {
                        'data': packages
                    }
                }
            }

        locks = []
        for env_spec_name, spec in locked_env_specs["env_specs"].items():
            env_spec_lock_id = uuid4().hex
            packages = []
            for platform, pkg_list in spec["packages"].items():
                for pkg in pkg_list:
                    packages.append(
                        create_pkg_for_wire(env_spec_lock_id, pkg, platform))
            hash_ = spec.get("env_spec_hash", "")
            locks.append(
                create_lock_for_wire(env_spec_name, env_spec_lock_id, hash_,
                                     packages))

        # hack because relationships are not being saved.
        relationships = locks[0].pop('relationships')
        resource = {'data': locks[0]}  # we assume only one spec
        self._json_api('env_spec_locks', method='POST', data=resource)

        resource['data']['relationships'] = relationships
        resource_id = resource['data']['id']
        self._json_api('env_spec_locks/{id}'.format(id=resource_id),
                       method='PATCH',
                       data=resource)


class WrapperFile(object):
    def __init__(self, fd, size):
        self.fd = fd
        self.size = size

    def __getattr__(self, attr):
        return getattr(self.fd, attr)

    def __len__(self):
        return self.size


def create_client(config_manager):
    try:
        site = config_manager.get_site()
    except ValueError as e:
        click.echo('Invalid site: {0}'.format(e))
        return
    config_manager.component = ['sites', site]
    token_store = TokenStore(config_manager.token_path)
    client = ApiClient(config_manager, token_store)
    client.check()
    return client


pass_client = click.make_pass_decorator(ApiClient)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2017 Continuum Analytics, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
import logging
from contextlib import contextmanager
from pprint import pprint

import click

from anaconda_enterprise.cli.config import ConfigurationError
from anaconda_enterprise.cli.internal import yaml


@contextmanager
def ignore_warnings():
    logging.disable(logging.WARNING)
    yield
    logging.disable(logging.NOTSET)


@click.group(name='config')
def main():
    """Manage configuration"""


@main.command()
@click.pass_obj
@click.argument('key')
@click.argument('value')
def set(config_manager, key, value):
    try:
        conf_yaml = yaml.load(value)
        new_conf = config_manager.new_config(key, conf_yaml)
    except ConfigurationError as e:
        raise click.ClickException("%s" % e)

    # this reads configuration file using load_yaml_file, which uses load_yaml_string
    # which runs the update function, which will print the warning on screen again
    # since it was shown during the loading of the config, so we ignore warnings.
    with ignore_warnings():
        user_conf = config_manager.get_user_config()

    new_conf = config_manager.merge_configs(user_conf, new_conf)
    config_manager.dump(new_conf)


@main.command()
@click.pass_obj
@click.argument('key')
def get(config_manager, key):
    try:
        value = config_manager[key]
    except KeyError:
        raise click.ClickException("Option '{key}' is not set".format(key=key))
    else:
        pprint(value)


@main.command()
@click.pass_obj
def view(config_manager):
    for path_ in config_manager.find_config_files():
        with open(path_) as fd:
            config_str = fd.read()
            click.echo('{}:'.format(path_))
            click.echo(config_str)
            click.echo()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2016 Continuum Analytics, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
import click

from anaconda_enterprise.cli.api import _list, create_client, pass_client


def _list_indexes(indexes, action=None):
    _list(('Name', 'Updated (UTC)'), ('name', 'updated'),
          [c['attributes'] for c in indexes],
          action=action,
          what='index',
          plural='es')


@click.group(name='pypi_indexes')
@click.pass_context
def main(ctx):
    """Manage pypi indexes"""
    ctx.obj = create_client(ctx.obj)


@main.command()
@pass_client
def list(client):
    _list_indexes(client.pypi_indexes())


@main.command()
@click.argument('name')
@pass_client
def create(client, name):
    _list_indexes([client.create_pypi_index(name)], action='created')


def _get_namespace(group, user, role):
    if group:
        return group, 'groups'
    elif user:
        return user, 'users'
    elif role:
        return role, 'roles'
    else:
        return None, None


@main.command()
@click.argument('pypi_index_name', metavar='pypi_index')
@click.option('--group', help='Group to share with')
@click.option('--user', help='User to share with')
@click.option('--role', help='Role to share with')
@click.option('--remove',
              is_flag=True,
              default=False,
              help='Remove share for a given user, group, or role.')
@click.option('--level',
              help='The level of permission to grant to the shared PyPI index',
              type=click.Choice([
                  'r',
                  'rw',
              ]),
              default='r')
@pass_client
def share(client, pypi_index_name, group, user, role, remove, level):
    pypi_index = client.get_pypi_index(pypi_index_name)
    name, namespace = _get_namespace(group, user, role)
    if namespace is not None:
        if remove:
            acl = client.get_pypi_index_acl(pypi_index, name, namespace)
            client.delete_pypi_index_acl(acl)
            click.echo('Permission removed.')
        else:
            client.create_pypi_index_acl(pypi_index, name, namespace, level)
            click.echo(
                'Shared PyPI index {} with {} {} with {} permission.'.format(
                    pypi_index_name, namespace, name, level))


@main.command()
@click.argument('pypi_index_name', metavar='pypi_index')
@pass_client
def delete(client, pypi_index_name):
    pypi_index = client.get_pypi_index(pypi_index_name)
    client.delete_pypi_index(pypi_index)
    _list_indexes([pypi_index], action='deleted')



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2017 Anaconda, Inc.
#
# All Rights Reserved.
# -----------------------------------------------------------------------------
import codecs
import collections
import email.parser
import hashlib
import json
import os.path
import re
import tarfile
import zipfile

from packaging.requirements import Requirement

from anaconda_enterprise.cli.internal import yaml

# 16 kB - amount of data to read from disk in one go
CHUNK_SIZE = 16384
# The license families that conda-build knows about
# https://github.com/conda/conda-build/blob/933a2178ffcd25ea3e96cc7e5d1904ec6cd9bab8/conda_build/license_family.py#L8
allowed_license_families = """
AGPL
LGPL
GPL3
GPL2
GPL
BSD
MIT
APACHE
PSF
CC
PUBLIC-DOMAIN
PROPRIETARY
OTHER
NONE
""".split()

# How to convert pep241 (Metadata 1.0) licenses to conda license families
pep241_licenses = {
    'BSD': 'BSD',
    'GNU GPL': 'GPL',
    'GNU LGPL': 'LGPL',
    'MIT': 'MIT',
    'public domain': 'PUBLIC-DOMAIN',
    'Python': 'PSF',
    'unknown': 'UNKNOWN',
}

# How to convert pypi license trove classifiers to conda license families
license_mappings = [
    ('License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
     'PUBLIC-DOMAIN'),
    ('License :: OSI Approved :: Apache Software License', 'APACHE'),
    ('License :: OSI Approved :: BSD License', 'BSD'),
    ('License :: OSI Approved :: GNU Affero General Public License v3',
     'AGPL'),
    ('License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
     'AGPL'),
    ('License :: OSI Approved :: GNU General Public License (GPL)', 'GPL'),
    ('License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
     'GPL2'),
    ('License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
     'GPL2'),
    ('License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
     'GPL3'),
    ('License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
     'GPL3'),
    ('License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
     'LGPL'),
    ('License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
     'LGPL'),
    ('License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
     'LGPL'),
    ('License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
     'LGPL'),
    ('License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
     'LGPL'),
    ('License :: OSI Approved :: MIT License', 'MIT'),
    ('License :: OSI Approved :: Python Software Foundation License', 'PSF'),
    ('License :: Other/Proprietary License', 'PROPRIETARY'),
    ('License :: Public Domain', 'PUBLIC-DOMAIN'),
]

# The fields defined in https://packaging.python.org/specifications/core-metadata/
# and prior PEPs as multiple use - meaning they can appear more than once.
multiple_use_fields = """\
platform
supported_platform
classifier
requires
requires_dist
provides_dist
obsoletes
obsoletes_dist
requires_external
project_url
provides
provides_extra""".split()

# The way that requirements in PKG-INFO/METADATA are marked as for "extras"
EXTRA_RE = re.compile('extra == [\'\"](.*)[\'\"]')

# https://www.python.org/dev/peps/pep-0440/#version-scheme
VERSION_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""

# https://www.python.org/dev/peps/pep-0508/#names
NAME_PATTERN = r"(?P<name>[A-Z0-9]|[A-Z0-9][A-Z0-9._-]*?[A-Z0-9])"

EXE_PATTERN = r"^" + NAME_PATTERN + r"-(?P<version>" + VERSION_PATTERN + r""")
    \.
    (?P<plat_name>win-amd64|win32)
    (?:-py(?P<python_version>[0-9]\.[0-9]))?
    \.exe$
"""

BDIST_WININST_RE = re.compile(EXE_PATTERN, re.VERBOSE | re.IGNORECASE)


def digest_file(filename):
    size = 0
    md5 = hashlib.md5()
    sha1 = hashlib.sha1()
    sha256 = hashlib.sha256()

    with open(filename, 'rb') as fd:
        for chunk in iter(lambda: fd.read(CHUNK_SIZE), b''):
            size += len(chunk)
            md5.update(chunk)
            sha1.update(chunk)
            sha256.update(chunk)

    return {
        'size': size,
        'md5': md5.hexdigest(),
        'sha1': sha1.hexdigest(),
        'sha256': sha256.hexdigest(),
    }


def parse_conda(filename, ignore_name_validation=False):
    """Open a conda package and extract information from the metada data

    Keywords:
    * filename -- path to the conda package
    * ingnore_name_validation -- if True this will not check that the name
      of the file is the same as declared on the metadata. This is useful
      for importing, since export might change the name of the package.
    """

    if not ignore_name_validation and not filename.endswith('.tar.bz2'):
        raise ValueError('Not a conda package (invalid extension)')

    with tarfile.open(filename, mode="r|bz2") as tf:
        found_info = False
        index_attributes = None
        about_attributes = None
        recipe_attributes = None
        for info in tf:
            is_info = info.name.startswith('info/')
            found_info = found_info or is_info
            if found_info and not is_info:
                break

            if info.name == "info/index.json":
                index = tf.extractfile(info)
                index_attributes = json.loads(index.read().decode())
            elif info.name == 'info/recipe.json':
                # recipe.json is deprecated and only packages built with older
                # versions of conda-build contain that file.
                recipe = tf.extractfile(info)
                recipe = json.loads(recipe.read().decode())
                about_attributes = recipe.pop('about', about_attributes)
            elif info.name == 'info/about.json':
                # about.json replaces recipe.json in newer versions of conda-build
                about = tf.extractfile(info)
                about_attributes = json.loads(about.read().decode())
            elif info.name in [
                    'info/recipe/meta.yaml.rendered', 'info/recipe/meta.yaml',
                    'info/meta.yaml'
            ]:
                try:
                    recipe = tf.extractfile(info)
                    recipe_attributes = yaml.load(recipe.read().decode())
                    if bool(recipe_attributes.get('app', {}).get('icon')):
                        recipe_attributes['has-icon'] = True

                except Exception:
                    # Pass yaml parsing exception.
                    # This exception doesn't affect uploading package
                    pass

        if index_attributes is None:
            raise ValueError('Not a conda package (missing info/index.json)')

        fn = '%(name)s-%(version)s-%(build)s.tar.bz2' % index_attributes
        if fn != os.path.basename(filename) and not ignore_name_validation:
            raise ValueError(
                'package has wrong filename (expected {0})'.format(fn))

    attributes = {
        'subdir': get_subdir(index_attributes),
        'version': index_attributes['version'],
        'filename': fn,
        'index-attributes': index_attributes,
        'about-attributes': about_attributes or {},
        'recipe-attributes': recipe_attributes or {},
    }
    attributes.update(digest_file(filename))
    return attributes


def get_subdir(index):
    try:
        return index['subdir']
    except KeyError:
        plat = index.get('platform')
        arch = index.get('arch')
        if arch is None:
            return 'noarch'
        intel_map = {'x86': '32', 'x86_64': '64'}
        return '%s-%s' % (plat, intel_map.get(arch, arch))


def canonicalize_license(license_name):
    # metadata 1.0 defines a fixed set of licenses:
    return pep241_licenses.get(license_name, 'OTHER')


def canonicalize_license_trove(classifiers):
    for expected_classifier, license in license_mappings:
        if expected_classifier in classifiers:
            return license
    return 'OTHER'


def make_attributes(metadata):
    assert not metadata.is_multipart(), "Unexpected multipart encoded PKG-INFO"
    json_metadata = {}
    # https://www.python.org/dev/peps/pep-0566/#json-compatible-metadata
    for key in metadata.keys():
        canonical_key = key.lower().replace('-', '_')
        if canonical_key == 'keywords':
            value = metadata.get(key).split()
        elif canonical_key in multiple_use_fields:
            value = metadata.get_all(key)
        else:
            value = metadata.get(key)
        json_metadata[canonical_key] = value

    payload = metadata.get_payload()
    if payload:
        json_metadata['description'] = payload

    metadata_version = metadata['Metadata-Version']
    if metadata_version == '1.0':
        license_family = canonicalize_license(metadata['License'])
    else:
        classifiers = json_metadata.get('classifier', [])
        license_family = canonicalize_license_trove(classifiers)

    about_attributes = {
        'license_family': license_family,
    }
    if 'Home-page' in metadata:
        about_attributes['home'] = metadata['Home-page']
    if 'Summary' in metadata:
        about_attributes['summary'] = metadata['Summary']

    index_attributes = {
        'name': json_metadata['name'],
    }
    if 'requires_python' in json_metadata:
        index_attributes['requires_python'] = json_metadata['requires_python']

    attributes = {
        'version': metadata['Version'],
        'index-attributes': index_attributes,
        'about-attributes': about_attributes,
        'pkg-info': json_metadata,
    }
    attributes.update(parse_metadata_requirements(json_metadata))
    return attributes


def safe_extra(extra):
    return re.sub('[^A-Za-z0-9.-]+', '_', extra.strip()).lower()


def safe_name(name):
    return re.sub('[^A-Za-z0-9.]+', '-', name)


def convert_requirements_to_attributes(common, extras):
    requirements = []
    extra_requirements = {}
    for req in common:
        requirements.append({
            'name':
            safe_name(req.name),
            'specs': [(spec.operator, spec.version) for spec in req.specifier],
            'value':
            str(req),
        })

    for extra in extras:
        extra_requirements[extra] = []
        for req in extras[extra]:
            extra_requirements[extra].append({
                'name':
                safe_name(req.name),
                'specs':
                [(spec.operator, spec.version) for spec in req.specifier],
                'value':
                str(req),
            })

    return {
        'requirements': requirements,
        'extra-requirements': extra_requirements,
    }


def yield_lines(strs):
    if isinstance(strs, (list, tuple)):
        for ss in strs:
            for s in yield_lines(ss):
                yield s
    else:
        assert isinstance(strs, str)
        for s in strs.splitlines():
            s = s.strip()
            if s and not s.startswith('#'):
                yield s


def split_sections(s):
    section = None
    content = []
    for line in yield_lines(s):
        if line.startswith("["):
            if line.endswith("]"):
                if section or content:
                    yield section, content
                section = line[1:-1].strip()
                content = []
            else:
                raise ValueError("Invalid section heading", line)
        else:
            content.append(line)
    yield section, content


def parse_requirements(strs):
    lines = iter(yield_lines(strs))
    for line in lines:
        # Drop comments -- a hash without a space may be in a URL.
        if ' #' in line:
            line = line[:line.find(' #')]
        if line.endswith('\\'):
            line = line[:-2].strip()
            try:
                line += next(lines)
            except StopIteration:
                return
        yield Requirement(line)


def parse_requirements_txt(requires_txt):
    sections = split_sections(requires_txt)
    common = []
    extras = collections.defaultdict(list)
    for section_name, content in sections:
        if section_name is None:
            common.extend(parse_requirements(content))
        else:
            extras[safe_extra(section_name)].extend(
                parse_requirements(content))

    return convert_requirements_to_attributes(common, extras)


def parse_metadata_requirements(metadata):
    if metadata['metadata_version'] in ('1.0', '1.1'):
        requires = metadata.get('requires', [])
    else:
        requires = metadata.get('requires_dist', [])

    def reqs_for_extra(extra):
        for req in reqs:
            if not req.marker or req.marker.evaluate({'extra': extra}):
                yield req

    reqs = list(parse_requirements(requires))
    common = frozenset(reqs_for_extra(None))
    extras = collections.defaultdict(list)

    for extra in metadata.get('provides_extra', []):
        extras[safe_extra(extra)].extend(
            frozenset(reqs_for_extra(extra)) - common)

    return convert_requirements_to_attributes(common, extras)


def parse_pypi(filename):
    if filename.endswith('.tar.gz') or filename.endswith('.tar.bz2'):
        attrs = parse_pypi_sdist(filename)
    elif filename.endswith('.whl'):
        attrs = parse_pypi_whl(filename)
    elif filename.endswith('.egg'):
        attrs = parse_pypi_egg(filename)
    elif filename.endswith('.zip'):
        attrs = parse_pypi_zip(filename)
    elif filename.endswith('.exe'):
        attrs = parse_pypi_exe(filename)
    else:
        _, ext = os.path.splitext(filename)
        raise ValueError(
            "Can not inspect pypi package with file extension %s" % ext)

    attrs['filename'] = os.path.basename(filename)
    attrs.update(digest_file(filename))
    return attrs


def parse_pypi_egg(filename):
    parser = email.parser.Parser()
    have_pkg_info = False
    attrs = {}

    with zipfile.ZipFile(filename, "r") as ar:
        for member in ar.infolist():
            if member.filename == 'EGG-INFO/PKG-INFO' and not have_pkg_info:
                with ar.open(member) as metadata:
                    with codecs.getreader('utf-8')(metadata) as wrapper:
                        fields = parser.parse(wrapper)
                attrs.update(make_attributes(fields))
                have_pkg_info = True
            elif member.filename.endswith('EGG-INFO/requires.txt'):
                requires_txt = ar.read(member)
                attrs.update(
                    parse_requirements_txt(requires_txt.decode('utf-8')))

    if not have_pkg_info:
        raise ValueError(
            'Could not parse egg package, missing EGG-INFO/PKG-INFO')
    return attrs


def parse_pypi_zip(filename):
    parser = email.parser.Parser()
    have_pkg_info = False
    attrs = {}

    with zipfile.ZipFile(filename, "r") as ar:
        for member in ar.infolist():
            if member.filename.endswith('/PKG-INFO') and not have_pkg_info:
                with ar.open(member) as metadata:
                    with codecs.getreader('utf-8')(metadata) as wrapper:
                        fields = parser.parse(wrapper)
                attrs.update(make_attributes(fields))
                have_pkg_info = True
            elif member.filename.endswith('.egg-info/requires.txt'):
                requires_txt = ar.read(member)
                attrs.update(
                    parse_requirements_txt(requires_txt.decode('utf-8')))

    if not have_pkg_info:
        raise ValueError('Could not parse sdist package, missing PKG-INFO')
    return attrs


def parse_pypi_sdist(filename):
    parser = email.parser.Parser()
    have_pkg_info = False
    attrs = {}

    with tarfile.open(filename, "r|*") as ar:
        for member in ar:
            if member.name.endswith('/PKG-INFO') and not have_pkg_info:
                metadata = ar.extractfile(member)
                with codecs.getreader('utf-8')(metadata) as wrapper:
                    fields = parser.parse(wrapper)
                attrs.update(make_attributes(fields))
                have_pkg_info = True
            elif member.name.endswith('.egg-info/requires.txt'):
                requires_txt = ar.extractfile(member).read()
                attrs.update(
                    parse_requirements_txt(requires_txt.decode('utf-8')))

    if not have_pkg_info:
        raise ValueError('Could not parse sdist package, missing PKG-INFO')
    return attrs


def parse_pypi_whl(filename):
    parser = email.parser.Parser()

    with zipfile.ZipFile(filename) as ar:
        for member in ar.infolist():
            if member.filename.endswith('.dist-info/METADATA'):
                with ar.open(member) as metadata:
                    with codecs.getreader('utf-8')(metadata) as wrapper:
                        fields = parser.parse(wrapper)
                return make_attributes(fields)

    raise ValueError('Missing .dist-info')


def parse_pypi_exe(filename):
    basename = os.path.basename(filename)
    match = BDIST_WININST_RE.match(basename)
    if match:
        name = match.group('name')
        version = match.group('version')
    else:
        raise ValueError(
            "Could not extract name & version from filename: {}".format(
                basename))

    return {
        'version': version,
        'index-attributes': {
            'name': name
        },
        'about-attributes': {},
        'pkg-info': {},
    }



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2016 Continuum Analytics, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
from click.testing import CliRunner
import pytest

from .base import VCR
from anaconda_enterprise.cli.main import group


@pytest.mark.integration
@VCR.use_cassette()
def test_upload_empty():
    runner = CliRunner()
    result = runner.invoke(group, ['login'], 'derek\nderek')
    assert result.exit_code == 0

    result = runner.invoke(
        group,
        ['upload', '--channel', 'test1'], )

    assert result.output == '''\
No packages uploaded.
'''
    assert result.exit_code == 0


@pytest.mark.integration
@VCR.use_cassette()
def test_upload_empty_with_default_channel(config):
    runner = CliRunner()
    runner.invoke(group, ['config', 'set', 'default_channel', 'test1'])
    result = runner.invoke(group, ['login'], 'derek\nderek')
    assert result.exit_code == 0

    result = runner.invoke(group, ['upload'])

    assert result.output == '''\
No packages uploaded.
'''
    assert result.exit_code == 0



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2016 Continuum Analytics, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
import os
from click.testing import CliRunner
import pytest

from .base import VCR
from anaconda_enterprise.cli.main import group
from anaconda_enterprise.cli import config


@pytest.mark.integration
@VCR.use_cassette()
def test_login():
    runner = CliRunner()
    result = runner.invoke(group, ['login'], 'derek\nderek')

    assert result.exception is None
    assert result.output == 'Username: derek\nPassword: \nLogged derek in!\n'
    assert result.exit_code == 0


@pytest.mark.integration
@VCR.use_cassette()
def test_login_failure():
    runner = CliRunner()
    result = runner.invoke(group, ['login'], 'derek\nfail')

    assert result.exit_code == 1


@pytest.mark.integration
@VCR.use_cassette()
def test_login_corrupted_token_store():
    with open(config.config.token_path, 'wb') as fd:
        fd.write(b'f/a/i/l')

    runner = CliRunner()
    result = runner.invoke(group, ['login'], 'derek\nderek')

    assert result.exception is None
    assert result.output == 'Username: derek\nPassword: \nLogged derek in!\n'
    assert result.exit_code == 0


@pytest.mark.integration
@VCR.use_cassette()
def test_logout_successful(tmpdir, monkeypatch):
    _config = config.config
    monkeypatch.setitem(_config.variables, _config.HOME_DIR, tmpdir)

    assert not os.path.exists(config.config.token_path)
    with open(config.config.token_path, 'wb') as fd:
        fd.write(b'secret_token_things')
    assert os.path.exists(config.config.token_path)

    runner = CliRunner()
    result = runner.invoke(group, ['logout'])

    assert result.exception is None
    assert result.exit_code == 0
    assert result.output == "User logged out successfully.\n"
    assert not os.path.exists(config.config.token_path)


@pytest.mark.integration
@VCR.use_cassette()
def test_logout_no_tokenfile(tmpdir, monkeypatch):
    _config = config.config
    monkeypatch.setitem(_config.variables, _config.HOME_DIR, tmpdir)

    assert not os.path.exists(config.config.token_path)

    runner = CliRunner()
    result = runner.invoke(group, ['logout'])

    assert result.output == "User already logged out: no active session found.\n"
    assert not os.path.exists(config.config.token_path)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2017 Anaconda, Inc.
#
# All Rights Reserved.
# -----------------------------------------------------------------------------
import os.path

import pytest

from anaconda_enterprise.cli import parse
from .base import FIXTURES

WHL = 'Flask-1.0.2-py2.py3-none-any.whl'
SDIST = 'Flask-1.0.2.tar.gz'


@pytest.mark.parametrize(['filename'], [
    (WHL, ),
    (SDIST, ),
])
def test_parse_pypi(filename):
    path = os.path.join(FIXTURES, filename)
    attrs = parse.parse_pypi(path)

    assert attrs['filename'] == filename
    assert attrs['version'] == '1.0.2'
    assert attrs['about-attributes']['license_family'] == 'BSD'
    assert attrs['index-attributes']['name'] == 'Flask'
    assert 'License :: OSI Approved :: BSD License' in attrs['pkg-info']['classifier']
    assert {'click', 'itsdangerous', 'Jinja2', 'Werkzeug'
            } == {req['name']
                  for req in attrs['requirements']}
    assert {
        'pytest', 'coverage', 'tox', 'sphinx', 'pallets-sphinx-themes', 'sphinxcontrib-log-cabinet'
    } == {req['name']
          for req in attrs['extra-requirements']['dev']}


@pytest.mark.parametrize(['filename', 'name', 'version'], [
    ('ipython-0.12.1.win-amd64.exe', 'ipython', '0.12.1'),
    ('simplejson-3.15.0.win-amd64-py2.7.exe', 'simplejson', '3.15.0'),
    ('simplejson-3.15.0.win32-py2.7.exe', 'simplejson', '3.15.0'),
    ('simple.json-3.15.0.win-amd64-py2.7.exe', 'simple.json', '3.15.0'),
    ('simple-json-3.15.0.win32-py2.7.exe', 'simple-json', '3.15.0'),
    ('simplejson-3.15.0-10.win32-py2.7.exe', 'simplejson', '3.15.0-10'),
    ('simple.json-3.15.0-10.win32-py2.7.exe', 'simple.json', '3.15.0-10'),
    ('simple-json-3.15.0-10.win32-py2.7.exe', 'simple-json', '3.15.0-10'),
])
def test_parse_pypi_exe(filename, name, version):
    attrs = parse.parse_pypi_exe(filename)

    assert attrs['index-attributes']['name'] == name
    assert attrs['version'] == version



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2016 Continuum Analytics, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
import click

from anaconda_enterprise.cli.api import _list, create_client, pass_client


@click.group(name='packages')
@click.pass_context
def main(ctx):
    """Manage conda packages"""
    client = create_client(ctx.obj)
    ctx.obj = client


def _list_files(files):
    _list(('Name', 'Platform', 'Owner'), ('filename', 'subdir', 'owner'),
          [f['attributes'] for f in files])


def _list_packages(packages):
    _list(('Name', 'Version', 'Channel'), ('name', 'version', 'channel-name'),
          [(p['attributes']['name'], v, p['attributes']['channel-name'])
           for p in sorted(packages, key=lambda p: p['attributes']['name'])
           for v in (p['attributes']['versions'] or [''])])


def _file_action(action, number, what='file'):
    click.echo('{} {}{} {}.'.format(number, what, '' if number == 1 else 's',
                                    action))


def _get_channel(client, channel):
    chan = client.get_channel(channel)
    if chan is None:
        raise click.ClickException(
            'Channel "{}" was not found'.format(channel))
    return chan


def _get_files(client, channel, package, list=True):
    client.get_channel(channel)
    pkg = client.get_package(channel, package)
    files = client.list_files(pkg)
    if list:
        _list_files(files)
    return pkg, files


@main.command()
@pass_client
def list(client):
    packages = client.packages()
    _list_packages(packages)


@main.command()
@click.argument('channel')
@click.argument('package')
@click.option('--file', help='platform/filename to delete')
@pass_client
def delete(client, channel, package, file):
    if file and '/' not in file:
        raise click.ClickException(
            "--file argument must be of the form platform/filename: {}".format(
                file))
    pkg, files = _get_files(client, channel, package, list=False)
    olen = len(files)
    if file:
        subdir, filename = file.split('/', 1)
        files = [
            f for f in files if f.get('attributes', {}).get('subdir') == subdir
            and f.get('attributes', {}).get('filename') == filename
        ]
        if not files:
            raise click.ClickException('File not found: {}/{}'.format(
                channel, file))
    else:
        subdir = filename = None
    nlen = len(files)
    if olen == nlen:
        msg = "This will delete all files related to {0}/{1}:"
    else:
        msg = "This will delete the following file related to {0}/{1}:"
    click.echo(msg.format(channel, package))
    _list_files(files)
    click.confirm("Do you want to continue?", abort=True)
    client.delete_package(pkg, subdir, filename)
    _file_action('deleted', nlen)


@main.command()
@click.argument('src_channel')
@click.argument('src_subdir')
@click.argument('src_filename')
@click.argument('dst_channel')
@pass_client
def copyfile(client, src_channel, src_subdir, src_filename, dst_channel):
    if src_channel == dst_channel:
        raise click.ClickException(
            'Source and destination channels are identical')
    _get_channel(client, src_channel)
    chan = _get_channel(client, dst_channel)
    client.copy_pkg_file(src_channel, src_subdir, src_filename, chan['id'])
    _file_action('copied', 1)


@main.command()
@click.argument('src_channel')
@click.argument('src_pkg')
@click.argument('dst_channel')
@pass_client
def copy(client, src_channel, src_pkg, dst_channel):
    if src_channel == dst_channel:
        raise click.ClickException(
            'Source and destination channels are identical')
    chan = _get_channel(client, dst_channel)
    _, files = _get_files(client, src_channel, src_pkg, list=True)
    client.copy_pkg(src_channel, src_pkg, chan['id'])
    _file_action('copied', len(files))


@main.command(name="list-files")
@click.argument('src_channel')
@click.argument('src_pkg')
@pass_client
def listfiles(client, src_channel, src_pkg):
    _, files = _get_files(client, src_channel, src_pkg, list=True)
    _file_action('found', len(files))
